using System;
using System.Collections.Generic;
using System.Text;
using MinorShift.Emuera.Sub;
using System.Text.RegularExpressions;
using MinorShift.Emuera.GameData.Variable;
using MinorShift.Emuera.GameData.Expression;
using MinorShift.Emuera.GameView;
using MinorShift.Emuera.GameData;
using MinorShift.Emuera.GameData.Function;

namespace MinorShift.Emuera.GameProc
{
	internal delegate void PrintWarning(string str, LogicalLine line, int level, bool isError);
	internal static class LogicalLineParser
	{
		readonly static Regex regexCom = new Regex("COM[0-9]+");
		readonly static Regex regexComAble = new Regex("COM_ABLE[0-9]+");
		readonly static Regex regexAblup = new Regex("ABLUP[0-9]+");
		private static bool isSystemLabel(string labelName)
		{
			switch (labelName)
			{
				case "EVENTFIRST":
				case "EVENTTRAIN":
				case "EVENTSHOP":
				case "EVENTBUY":
				case "EVENTCOM":
				case "EVENTTURNEND":
				case "EVENTCOMEND":
				case "EVENTEND":
				case "SHOW_STATUS":
				case "SHOW_USERCOM":
				case "USERCOM":
				case "SOURCE_CHECK":
				case "SHOW_JUEL":
				case "SHOW_ABLUP_SELECT":
				case "USERABLUP":
				case "SHOW_SHOP":
				case "SAVEINFO":
				case "USERSHOP":

				case "EVENTLOAD":
				case "TITLE_LOADGAME":
				case "SYSTEM_AUTOSAVE":
				case "SYSTEM_TITLE":
					return true;
			}

			if (labelName.StartsWith("COM"))
			{
				if (regexCom.IsMatch(labelName))
					return true;
				if (regexComAble.IsMatch(labelName))
					return true;
			}
			if (labelName.StartsWith("ABLUP"))
				if (regexAblup.IsMatch(labelName))
					return true;
			return false;
		}

		readonly static char[] badSymbolAsLabelname = new char[]
		{
		    '+', '-', '*', '/', '%', '=', '!', '<', '>', '|', '&', '^', '~',
		    ' ', '@', '\t' ,
		    '\"','(', ')', '{', '[', '}', ']', ',', '.', ':',
		    '\\', '@', '$', '#', '?',
		    //'_'OK
		};

		public static LogicalLine ParseLine(string str, EmueraConsole console)
		{
			ScriptPosition position = new ScriptPosition(str);
			StringStream stream = new StringStream(str);
			return ParseLine(stream, position, console);
		}

		public static LogicalLine ParseLine(StringStream stream, ScriptPosition position, EmueraConsole console)
		{
			int lineNo = position.LineNo;
			string rowLine = stream.Substring();
			string s1 = null;
			string s2 = null;
			string s3 = null;
			string errMes = "";
			TokenReader.SkipWhiteSpace(stream);//擪̃zCgXy[Xǂݔ΂
			if (stream.EOS)
				return null;
			//Rgsǂ͂ɗOɔ肵Ă
			//x
			if ((stream.Current == '$') || (stream.Current == '@'))
			{
				bool isFunction = (stream.Current == '@');
				stream.ShiftNext();//@$

				string labelName = TokenReader.ReadStringEndWith(stream, new char[]{',','('});
				List<VariableToken> args = new List<VariableToken>();
				labelName = labelName.Trim();
				if (Config.Instance.IgnoreCase)
					labelName = labelName.ToUpper();
				if (labelName.Length == 0)
				{
					errMes = "x܂";
					goto err;
				}
				//////Eramakerł̓ubNXgɂL͑Sʂ̂Ō݊mۂ̂߂ȂƂ
				////ϐɎgȂgLOCAL@֐ō̂ŕB
				////eramakerł͐ɓEmuerałʏ͖Ȃ삷̂ŌxLv0ɒቺB
				//1.721 LT|[gȂɕύXB
				if (labelName.IndexOfAny(badSymbolAsLabelname) >= 0)
					console.PrintWarning("x" + labelName + "\"_\"ȊŐL܂܂Ă܂B", position, 1);
                //else if (firstTokenType == TokenType.Numeric)
                else if (char.IsNumber(labelName[0]))
					console.PrintWarning("x" + labelName + "pn܂Ă܂", position, 0);


				if (!stream.EOS)
				{
					if (!isFunction)
					{
						console.PrintWarning("$Ŏn܂郉xɈݒ肳Ă܂", position, 1);
					}
					else
					{
						while (!stream.EOS)
						{
							if ((stream.Current != ',')&&(stream.Current != '('))
							{
								errMes = "̏ԈĂ܂";
								goto err;
							}
							bool startWithBracket = stream.Current == '(';
							if (!startWithBracket)
								stream.ShiftNext();
							IOperandTerm[] opArgs = ExpressionParser.ReduceArguments(stream, startWithBracket);
							for (int i = 0; i < opArgs.Length; i++)
							{
								if (opArgs[i] == null)
								{
									errMes = "֐`̈͏ȗł܂";
									goto err;
								}
								if (!(opArgs[i] is VariableToken))
								{
									errMes = "֐`̈ɂ͕ϐw肵Ă";
									goto err;
								}
								VariableToken vToken = (VariableToken)opArgs[i];
								if (vToken.IsConstant)
								{
									errMes = "֐`̈ɂ͑\ȕϐw肵Ă";
									goto err;
								}
								args.Add(vToken);
							}

							//stream.ShiftNext();
							//TokenReader.SkipWhiteSpace(stream);
							//VariableToken vars = VariableParser.ReduceVariable(stream);
							//args.Add(vars);
							//TokenReader.SkipWhiteSpace(stream);
						}
					}
				}
				if (isFunction)
				{
					FunctionLabelLine labelLine = new FunctionLabelLine(position, labelName, args);
					if (isSystemLabel(labelName))
						labelLine.Depth = 0;
					return labelLine;
				}
				else
				{
					return new GotoLabelLine(position, labelName);
				}
			}
            //OuCNgAfNgs
            if (stream.Current == '+')
            {
                stream.ShiftNext();
                if (stream.Current != '+')
                {
                    errMes = "s\'+\'n܂Ă܂ACNgł͂܂";
                    goto err;
                }
                stream.ShiftNext();
                TokenReader.SkipWhiteSpace(stream);
                s1 = TokenReader.ReadIdentiferWithIndex(stream);
                if (!VariableParser.IsVariable(s1))
                {
                    errMes = "CNg̑Ώۂϐł͂܂";
                    goto err;
                }
                if (!VariableParser.IsInteger(s1))
                {
                    errMes = "lϐȊO̕ϐCNg悤Ƃ܂";
                    goto err;
                }
                if (VariableIdentifier.GetVariableId(s1.Split(':')[0]).Readonly)
                {
                    errMes = "ύXsȒlɑ΂ăCNgs悤Ƃ܂";
                    goto err;
                }
                BuiltInFunctionCode f = BuiltInFunctionCode.SET;
                string r = s1 + " + 1";
                return new InstructionLine(position, f, s1, r);
            }
            else if (stream.Current == '-')
            {
                stream.ShiftNext();
                if (stream.Current != '-')
                {
                    errMes = "s\'-\'n܂Ă܂AfNgł͂܂";
                    goto err;
                }
                stream.ShiftNext();
                TokenReader.SkipWhiteSpace(stream);
                s1 = TokenReader.ReadIdentiferWithIndex(stream);
                if (!VariableParser.IsVariable(s1))
                {
                    errMes = "fNg̑Ώۂϐł͂܂";
                    goto err;
                }
                if (!VariableParser.IsInteger(s1))
                {
                    errMes = "lϐȊO̕ϐfNg悤Ƃ܂";
                    goto err;
                }
                if (VariableIdentifier.GetVariableId(s1.Split(':')[0]).Readonly)
                {
                    errMes = "ύXsȒlɑ΂ăfNgs悤Ƃ܂";
                    goto err;
                }
                BuiltInFunctionCode f = BuiltInFunctionCode.SET;
                string r = s1 + " - 1";
                return new InstructionLine(position, f, s1, r);
            }

			s1 = TokenReader.ReadIdentiferWithIndex(stream);
			BuiltInFunctionCode func = BuiltInFunctionManager.ToBuiltInFunction(s1);
			//ߕ
			if (func != BuiltInFunctionCode.__NULL__)//֐
			{
				if (stream.EOS) //̖֐
					return new InstructionLine(position, func, null);
				stream.ShiftNext();
				s2 = rowLine.Substring(stream.CurrentPosition, rowLine.Length - stream.CurrentPosition);
				return new InstructionLine(position, func, s2);
			}
			//xł߂łȂ̂ő̂͂BϐŎn܂ĂȂ΂B
			if (!VariableParser.IsVariable(s1))
			{
				errMes = "xEߕÊƂ߂łȂsł";
				goto err;
			}
			if (VariableParser.IsInteger(s1))
				func = BuiltInFunctionCode.SET;
			else
				func = BuiltInFunctionCode.SETS;

			TokenReader.SkipWhiteSpace(stream);//̒P܂œǂݔ΂
			//"ϐ@Zq@"Ƃ`̂͂B
			if ((stream.EOS) || (TokenReader.GetNextTokenType(stream) != TokenType.Operator))
			{
				errMes = "sϐŎn܂Ă܂Zq܂";
				goto err;
			}
			s2 = TokenReader.ReadOperatorString(stream);
			if (s2 == "==")
			{
				if (console != null)
					console.PrintWarning("Zq\"==\"gĂ܂", position, 0);
				s2 = "=";//"=="Ɏĝ͖{͂\gĂ̂ŎdlɂB
			}
			if (func == BuiltInFunctionCode.SET)
				TokenReader.SkipWhiteSpace(stream);//̑Ȃ玟̒P܂œǂݔ΂
			else if (stream.Current == ' ')
				TokenReader.SkipHalfSpace(stream);//̑Ȃ甼pXy[Xǂݔ΂

			if (stream.EOS)
				s3 = "";
			else
				s3 = rowLine.Substring(stream.CurrentPosition, rowLine.Length - stream.CurrentPosition);


            //if (func == BuiltInFunctionCode.SETS)
            //{
            //    if (s2 != "=")
            //    {
            //        errMes = "ϐ̑ɂ͕Zq͎g܂";
            //        goto err;
            //    }
            //}
			string right = null;
            if (func == BuiltInFunctionCode.SETS)
            {
                //right = s3;
                right = OperatorManager.ReduceAssignmentOperatorString(s2, s1, s3);
                if (right == null)
                {
                    errMes = "ɎgpłȂZqg܂";
                    goto err;
                }
            }
            else
            {
                right = OperatorManager.ReduceAssignmentOperator(s2, s1, s3);
                if (right == null)
                {
                    errMes = "ɎgpłȂZqg܂";
                    goto err;
                }
            }

			return new InstructionLine(position, func, s1, right);
		err:
			return new InvalidLine(position, errMes);
		}

		private static void printWarning(string str, LogicalLine line, int level, bool isError)
		{
			if (isError)
			{
				line.IsError = true;
				line.ErrMes = str;
			}
			if (printWarningLoad == null)
				return;
			printWarningLoad(str, line, level, isError);
		}

		static PrintWarning printWarningLoad = null;
		public static bool SetArgumentTo(InstructionLine line, PrintWarning printW)
		{
			if (line == null)
				return false;
			if (line.Argument != null)
				return true;
			if (line.IsError)
				return false;
			printWarningLoad = printW;
			BuiltInFunctionCode argType = BuiltInFunctionCode.__ARG_MASK__ & line.Function;
			string rowStr = line.ArgumentStr;
			StringStream st = new StringStream(rowStr);
			Argument arg = null;
			IOperandTerm term = null;
			VariableToken varToken = null;
			StringForm strForm = null;
			try
			{
				if ((line.Function & BuiltInFunctionCode.__ARG_METHOD__) == BuiltInFunctionCode.__ARG_METHOD__)
				{
					IOperandTerm[] args = ExpressionParser.ReduceArguments(st,false);
					string code = line.Function.ToString();
					IOperandTerm mTerm = FunctionMethodCreator.GetFunctionMethod(null, code, args);
					arg = new FunctionArgument(mTerm);
				}
				else
				{
					switch (argType)
					{
						case BuiltInFunctionCode.__ARG_VOID__:
							if ((rowStr != null) && (rowStr.Length != 0) && (rowStr.Trim().Length != 0))
								printWarning("svȖ\"" + line.Function.ToString() + "\"Ɉw肳Ă܂", line, 1, false);
							arg = new VoidArgument();
							break;
						case BuiltInFunctionCode.__ARG_INT_EXPRESSION__:
							term = ExpressionParser.ReduceIntegerTerm(st);
							arg = new ExpressionArgument(term);
							break;
						case BuiltInFunctionCode.__ARG_STR_EXPRESSION__:
							term = ExpressionParser.ReduceStringTerm(st);
							arg = new ExpressionArgument(term);
							break;
						case BuiltInFunctionCode.__ARG_EXPRESSION__:
							term = ExpressionParser.ReduceExpressionTerm(st);
							arg = new ExpressionArgument(term);
							break;
						case BuiltInFunctionCode.__ARG_STR__:
							if (rowStr == null)
								rowStr = "";
							arg = new ExpressionArgument(new SingleTerm(rowStr));
							break;
						case BuiltInFunctionCode.__ARG_FORM_STR__:
							if (rowStr == null)
								rowStr = "";
							strForm = new StringForm(rowStr);
							if (Config.Instance.ReduceFormattedStringOnLoad)
								strForm.Reduce();
							arg = new ExpressionArgument(new StringFormTerm(strForm));
							break;
						case BuiltInFunctionCode.__ARG_SP_PRINTV__:
							{
								List<IOperandTerm> list = new List<IOperandTerm>();
								while (!st.EOS)
								{
									TokenReader.SkipWhiteSpace(st);
									if (st.Current == '\'')
									{
										st.ShiftNext();
										if (!st.EOS && st.Current != ',')
											list.Add(new SingleTerm(TokenReader.ReadStringEndWith(st, null)));
									}
									else
										list.Add(ExpressionParser.ReduceIntegerTerm(st));
									TokenReader.SkipWhiteSpace(st);
									if (st.Current == ',')
										st.ShiftNext();
									else
										break;
								}
								arg = new SpPrintVArgument(list);
							}
							break;
						case BuiltInFunctionCode.__ARG_SP_TIMES__:
							TokenReader.SkipWhiteSpace(st);
							varToken = VariableParser.ReduceVariable(st);
							if (varToken == null)
								printWarning("\"" + line.Function.ToString() + "\"̑PFł܂", line, 2, true);
							else if (varToken.IsString)
								printWarning("\"" + line.Function.ToString() + "\"̑P𕶎ϐɂ邱Ƃ͂ł܂", line, 2, true);
							else if (varToken.Identifier.Readonly)
								printWarning("\"" + line.Function.ToString() + "\"̑PɕύXłȂϐw肷邱Ƃ͂ł܂", line, 2, true);

							TokenReader.SkipWhiteSpace(st);
							if (st.Current != ',')
							{
								if (st.EOS)
									printWarning("\"" + line.Function.ToString() + "\"̈܂", line, 2, true);
								else
									printWarning("\"" + line.Function.ToString() + "\"̏ԈĂ܂", line, 2, true);
								break;
							}
							st.ShiftNext();
							TokenReader.SkipWhiteSpace(st);
							double d = 0.0;
							if ((TokenReader.GetNextTokenType(st) != TokenType.Numeric) && (st.Current != '-') && (st.Current != '.'))
								printWarning("\"" + line.Function.ToString() + "\"̑Qlł͂܂i0Ɖ߂܂j", line, 1, false);
							else
							{
								try
								{
									d = ExpressionParser.ReduceDoubleTerm(st);
								}
								catch
								{
									printWarning("\"" + line.Function.ToString() + "\"̑Qlł͂܂i0Ɖ߂܂j", line, 1, false);
								}
							}
							arg = new SpTimesArgument(varToken, d);
							break;
						case BuiltInFunctionCode.__ARG_SP_BAR__:
							{
								TokenReader.SkipWhiteSpace(st);
								IOperandTerm value = ExpressionParser.ReduceIntegerTerm(st);
								TokenReader.SkipWhiteSpace(st);
								if (st.Current != ',')
								{
									printWarning("\"" + line.Function.ToString() + "\"̈܂", line, 2, true);
									break;
								}
								st.ShiftNext();
								IOperandTerm max = ExpressionParser.ReduceIntegerTerm(st);
								TokenReader.SkipWhiteSpace(st);
								if (st.Current != ',')
								{
									printWarning("\"" + line.Function.ToString() + "\"̈܂", line, 2, true);
									break;
								}
								st.ShiftNext();
								IOperandTerm length = ExpressionParser.ReduceIntegerTerm(st);
								TokenReader.SkipWhiteSpace(st);
								if (st.Current == ',')
									printWarning("\"" + line.Function.ToString() + "\"̈܂", line, 1, false);
								arg = new SpBarArgument(value, max, length);
							}
							break;
						case BuiltInFunctionCode.__ARG_SP_SET__:
							varToken = VariableParser.ReduceVariable(new StringStream(line.AssignmentDestStr));
							term = ExpressionParser.ReduceIntegerTerm(st);
							arg = new SpSetArgument(varToken, term);
							break;

						case BuiltInFunctionCode.__ARG_SP_SETS__:
							varToken = VariableParser.ReduceVariable(new StringStream(line.AssignmentDestStr));
							strForm = new StringForm(rowStr);
							if (Config.Instance.ReduceFormattedStringOnLoad)
								strForm.Reduce();
							arg = new SpSetsArgument(varToken, strForm);
							break;

						//case BuiltInFunctionCode.__ARG_SP_SUBSTRING__:
						//    {
						//        TokenReader.SkipWhiteSpace(st);
						//        term = ExpressionParser.ReduceStringTerm(st);
						//        if (term == null)
						//            printWarning("\"" + line.Function.ToString() + "\"̑PFł܂", line, 2, true);
						//        TokenReader.SkipWhiteSpace(st);
						//        if (st.Current != ',')
						//        {
						//            if (st.EOS)
						//                printWarning("\"" + line.Function.ToString() + "\"̈܂", line, 2, true);
						//            else
						//                printWarning("\"" + line.Function.ToString() + "\"̏ԈĂ܂", line, 2, true);
						//            break;
						//        }
						//        st.ShiftNext();
						//        TokenReader.SkipWhiteSpace(st);
						//        IOperandTerm start = ExpressionParser.ReduceIntegerTerm(st);
						//        TokenReader.SkipWhiteSpace(st);
						//        if (st.Current != ',')
						//        {
						//            printWarning("\"" + line.Function.ToString() + "\"̈܂", line, 2, true);
						//            break;
						//        }
						//        st.ShiftNext();
						//        TokenReader.SkipWhiteSpace(st);
						//        IOperandTerm length = ExpressionParser.ReduceIntegerTerm(st);
						//        TokenReader.SkipWhiteSpace(st);
						//        if (st.Current == ',')
						//            printWarning("\"" + line.Function.ToString() + "\"̈܂", line, 1, false);

						//        arg = new SpSubstringArgument(term, start, length);
						//    }
						//    break;
						case BuiltInFunctionCode.__ARG_SP_SWAP__:
							{
								TokenReader.SkipWhiteSpace(st);
								IOperandTerm x = ExpressionParser.ReduceIntegerTerm(st);
								TokenReader.SkipWhiteSpace(st);
								if (st.Current != ',')
								{
									printWarning("\"" + line.Function.ToString() + "\"̈܂", line, 2, true);
									break;
								}
								st.ShiftNext();
								IOperandTerm y = ExpressionParser.ReduceIntegerTerm(st);
								TokenReader.SkipWhiteSpace(st);
								if (st.Current == ',')
									printWarning("\"" + line.Function.ToString() + "\"̈܂", line, 1, false);
								arg = new SpSwapCharaArgument(x, y);
							}
							break;
						case BuiltInFunctionCode.__ARG_SP_VAR__:
							{
								TokenReader.SkipWhiteSpace(st);

								string idStr = TokenReader.ReadSingleIdentifer(st);
								VariableIdentifier id = VariableIdentifier.GetVariableId(idStr);
								if (id == null)
									printWarning("\"" + line.Function.ToString() + "\"̑PFł܂", line, 2, true);
								else if ((!id.IsArray1D && !id.IsArray2D && !id.IsArray3D) || (id.Code == VariableCode.RAND))
									printWarning("\"" + line.Function.ToString() + "\"̑PɔzłȂϐw肷邱Ƃ͂ł܂", line, 2, true);
								arg = new SpVarsizeArgument(id);
							}
							break;
						case BuiltInFunctionCode.__ARG_SP_SAVEDATA__:
							{
								TokenReader.SkipWhiteSpace(st);
								IOperandTerm target = ExpressionParser.ReduceIntegerTerm(st);
								TokenReader.SkipWhiteSpace(st);
								if (st.Current != ',')
								{
									if (st.EOS)
										printWarning("\"" + line.Function.ToString() + "\"̈܂", line, 2, true);
									else
										printWarning("\"" + line.Function.ToString() + "\"̏ԈĂ܂", line, 2, true);
									break;
								}
								st.ShiftNext();
								TokenReader.SkipWhiteSpace(st);
								term = ExpressionParser.ReduceStringTerm(st);
								if (term == null)
									printWarning("\"" + line.Function.ToString() + "\"̑QFł܂", line, 2, true);
								TokenReader.SkipWhiteSpace(st);
								if (!st.EOS)
								{
									if (st.Current == ',')
										printWarning("\"" + line.Function.ToString() + "\"̈܂", line, 1, true);
									else
										printWarning("\"" + line.Function.ToString() + "\"̏ԈĂ܂", line, 1, true);
									break;
								}
								arg = new SpSaveDataArgument(target, term);
								break;
							}
						case BuiltInFunctionCode.__ARG_SP_TINPUTS__:
							TokenReader.SkipWhiteSpace(st);
							IOperandTerm time = ExpressionParser.ReduceIntegerTerm(st);
							TokenReader.SkipWhiteSpace(st);
							if (st.Current != ',')
							{
								if (st.EOS)
									printWarning("\"" + line.Function.ToString() + "\"̈܂", line, 2, true);
								else
									printWarning("\"" + line.Function.ToString() + "\"̏ԈĂ܂", line, 2, true);
								break;
							}
							st.ShiftNext();
							TokenReader.SkipWhiteSpace(st);
							term = ExpressionParser.ReduceStringTerm(st);
							if (term == null)
								printWarning("\"" + line.Function.ToString() + "\"̑QFł܂", line, 2, true);
							TokenReader.SkipWhiteSpace(st);
							if (st.Current != ',')
							{
								printWarning("\"" + line.Function.ToString() + "\"̈܂", line, 2, true);
								break;
							}
							st.ShiftNext();
							IOperandTerm disp = ExpressionParser.ReduceIntegerTerm(st);
							TokenReader.SkipWhiteSpace(st);
							if (st.Current == ',')
								printWarning("\"" + line.Function.ToString() + "\"̈܂", line, 1, false);
							arg = new SpTInputsArgument(time, term, disp);
							break;
						case BuiltInFunctionCode.__ARG_SP_SORTCHARA__:
							{
								varToken = new VariableToken(VariableIdentifier.GetVariableId("NO"), null, null, null);
								SortOrder order = SortOrder.ASCENDING;
								TokenReader.SkipWhiteSpace(st);
								if (!st.EOS && !st.CurrentEqualTo("FORWARD") && !st.CurrentEqualTo("BACK"))
								{
									varToken = VariableParser.ReduceVariable(st);
									if (varToken == null)
										printWarning("\"" + line.Function.ToString() + "\"̑PFł܂", line, 2, true);
									else if (!varToken.Identifier.IsCharacterData)
										printWarning("\"" + line.Function.ToString() + "\"̑P̓LN^ϐłȂ΂Ȃ܂", line, 2, true);
									TokenReader.SkipWhiteSpace(st);
									if (st.Current != ',' && !st.EOS)
									{
										printWarning("\"" + line.Function.ToString() + "\"̏ԈĂ܂", line, 2, true);
									}
								}
								st.ShiftNext();
								TokenReader.SkipWhiteSpace(st);
								if (!st.EOS)
								{
									if (st.CurrentEqualTo("FORWARD"))
										order = SortOrder.ASCENDING;
									else if (st.CurrentEqualTo("BACK"))
										order = SortOrder.DESENDING;
									else
										printWarning("\"" + line.Function.ToString() + "\"̏ԈĂ܂", line, 2, true);
								}

								arg = new SpSortcharaArgument(varToken, order);
								break;
							}

						case BuiltInFunctionCode.__ARG_SP_CALL__:
							{
								TokenReader.SkipWhiteSpace(st);
								string str = TokenReader.ReadStringEndWith(st, new char[] { ',', '(' });
								str = str.Trim();
								if (str.Length == 0)
								{
									printWarning(line.Function.ToString() + "߂Ɋ֐w肳Ă܂", line, 2, true);
									break;
								}
								TokenReader.SkipWhiteSpace(st);
								IOperandTerm[] args = null;
								char start = st.Current;
								if (st.EOS)
									args = new IOperandTerm[0];
								else if ((start == '(') || (st.Current == ','))
								{
									bool startWithBracket = st.Current == '(';
									if(!startWithBracket)
										st.ShiftNext();
									args = ExpressionParser.ReduceArguments(st,startWithBracket);
									if (!st.EOS)
										printWarning("\"" + line.Function.ToString() + "\"̏ԈĂ܂", line, 1, false);
								}
								else
								{
									printWarning("\"" + line.Function.ToString() + "\"̏ԈĂ܂", line, 2, true);
									break;
								}
								arg = new SpCallArgment(str, args);
							}
							break;
						case BuiltInFunctionCode.__ARG_SP_CALLFORM__:
							{
								TokenReader.SkipWhiteSpace(st);
								List<IOperandTerm> terms = new List<IOperandTerm>();
								string str = TokenReader.ReadFormStringEndWith(st, new char[] { ',', '(' });
								str = str.Trim();
								if (str.Length == 0)
								{
									printWarning(line.Function.ToString() + "߂Ɋ֐w肳Ă܂", line, 2, true);
									break;
								}
								strForm = new StringForm(str);
								IOperandTerm[] args = null;
								char start = st.Current;
								if (st.EOS)
								{
									args = new IOperandTerm[0];
								}
								else if ((start == '(') || (st.Current == ','))
								{
									bool startWithBracket = st.Current == '(';
									if (!startWithBracket)
										st.ShiftNext();
									args = ExpressionParser.ReduceArguments(st,startWithBracket);
									if (!st.EOS)
										printWarning("\"" + line.Function.ToString() + "\"̏ԈĂ܂", line, 1, false);
								}
								else
								{
									printWarning("\"" + line.Function.ToString() + "\"̏ԈĂ܂", line, 2, true);
									break;
								}
								arg = new SpCallformArgment(strForm, args);
							}
							break;
						case BuiltInFunctionCode.__ARG_SP_FOR_NEXT__:
							{
								TokenReader.SkipWhiteSpace(st);
								varToken = VariableParser.ReduceVariable(st);
								IOperandTerm start = null;
								IOperandTerm end = null;
								IOperandTerm step = null;
								if (varToken == null)
								{ printWarning("\"" + line.Function.ToString() + "\"̑PFł܂", line, 2, true); break; }
								else if (!varToken.Identifier.IsInteger)
								{ printWarning("\"" + line.Function.ToString() + "\"̑P͐l^ϐłȂ΂Ȃ܂", line, 2, true); break; }
								else if (varToken.Identifier.IsCharacterData)
								{ printWarning("\"" + line.Function.ToString() + "\"̑PɃLN^ϐw肷邱Ƃ͂ł܂", line, 2, true); break; }
								else if (varToken.Identifier.Readonly)
								{ printWarning("\"" + line.Function.ToString() + "\"̑P͑\ȕϐłȂ΂Ȃ܂", line, 2, true); break; }
								TokenReader.SkipWhiteSpace(st);
								if (st.EOS)
								{ printWarning("\"" + line.Function.ToString() + "\"̈܂", line, 2, true); break; }
								else if (st.Current != ',')
								{ printWarning("\"" + line.Function.ToString() + "\"̏ԈĂ܂", line, 2, true); break; }
								st.ShiftNext();
								if (st.Current == ',')
									start = new SingleTerm(0);
								else
									start = ExpressionParser.ReduceIntegerTerm(st);
								TokenReader.SkipWhiteSpace(st);
								if (st.EOS)
								{ printWarning("\"" + line.Function.ToString() + "\"̈܂", line, 2, true); break; }
								else if (st.Current != ',')
								{ printWarning("\"" + line.Function.ToString() + "\"̏ԈĂ܂", line, 2, true); break; }
								st.ShiftNext();
								end = ExpressionParser.ReduceIntegerTerm(st);
								TokenReader.SkipWhiteSpace(st);
								if (st.EOS)
									step = new SingleTerm(1);
								else if (st.Current == ',')
								{
									st.ShiftNext();
									step = ExpressionParser.ReduceIntegerTerm(st);
								}
								else
								{ printWarning("\"" + line.Function.ToString() + "\"̏ԈĂ܂", line, 2, true); break; }
								arg = new SpForNextArgment(varToken, start, end, step);
								TokenReader.SkipWhiteSpace(st);
								break;
							}

						case BuiltInFunctionCode.__ARG_SP_POWER__:
							{
								TokenReader.SkipWhiteSpace(st);
								varToken = VariableParser.ReduceVariable(st);
								if (varToken == null)
									printWarning("\"" + line.Function.ToString() + "\"̑PFł܂", line, 2, true);
								else if (varToken.IsString)
									printWarning("\"" + line.Function.ToString() + "\"̑P𕶎ϐɂ邱Ƃ͂ł܂", line, 2, true);
								else if (varToken.Identifier.Readonly)
									printWarning("\"" + line.Function.ToString() + "\"̑PɕύXłȂϐw肷邱Ƃ͂ł܂", line, 2, true);

								TokenReader.SkipWhiteSpace(st);
								if (st.Current != ',')
								{
									printWarning("\"" + line.Function.ToString() + "\"̈܂", line, 2, true);
									break;
								}
								st.ShiftNext();
								TokenReader.SkipWhiteSpace(st);
								IOperandTerm x = ExpressionParser.ReduceIntegerTerm(st);
								TokenReader.SkipWhiteSpace(st);
								if (st.Current != ',')
								{
									printWarning("\"" + line.Function.ToString() + "\"̈܂", line, 2, true);
									break;
								}
								st.ShiftNext();
								TokenReader.SkipWhiteSpace(st);
								IOperandTerm y = ExpressionParser.ReduceIntegerTerm(st);
								TokenReader.SkipWhiteSpace(st);
								if (st.Current == ',')
									printWarning("\"" + line.Function.ToString() + "\"̈܂", line, 1, false);

								arg = new SpPowerArgument(varToken, x, y);
								break;
							}
						case BuiltInFunctionCode.__ARG_SP_SWAPVAR__:
							{
								TokenReader.SkipWhiteSpace(st);
								VariableToken x = VariableParser.ReduceVariable(st);
								TokenReader.SkipWhiteSpace(st);
								if (st.Current != ',')
								{
									printWarning("\"" + line.Function.ToString() + "\"̈܂", line, 2, true);
									break;
								}
								st.ShiftNext();
								TokenReader.SkipWhiteSpace(st);
								VariableToken y = VariableParser.ReduceVariable(st);
								TokenReader.SkipWhiteSpace(st);
								if (st.Current == ',')
									printWarning("\"" + line.Function.ToString() + "\"̈܂", line, 1, false);

								if (x.GetOperandType() != y.GetOperandType())
								{
									printWarning("̌^قȂ܂", line, 2, true);
									break;
								}


								arg = new SpSwapVarArgument(x, y);
								break;
							}
						//case BuiltInFunctionCode.__ARG_SP_STRFIND__:
						//    {
						//        TokenReader.SkipWhiteSpace(st);
						//        IOperandTerm target = ExpressionParser.ReduceStringTerm(st);
						//        TokenReader.SkipWhiteSpace(st);
						//        if (st.Current != ',')
						//        {
						//            if (st.EOS)
						//                printWarning("\"" + line.Function.ToString() + "\"̈܂", line, 2, true);
						//            else
						//                printWarning("\"" + line.Function.ToString() + "\"̏ԈĂ܂", line, 2, true);
						//            break;
						//        }
						//        st.ShiftNext();
						//        TokenReader.SkipWhiteSpace(st);
						//        IOperandTerm word = ExpressionParser.ReduceStringTerm(st);
						//        if (word == null)
						//        {
						//            printWarning("\"" + line.Function.ToString() + "\"̑QFł܂", line, 2, true);
						//            break;
						//        }
						//        TokenReader.SkipWhiteSpace(st);
						//        IOperandTerm start = new SingleTerm(0);
						//        if (!st.EOS)
						//        {
						//            if (st.Current == ',')
						//            {
						//                start = ExpressionParser.ReduceIntegerTerm(st);
						//                if (start == null)
						//                {
						//                    printWarning("\"" + line.Function.ToString() + "\"̑RFł܂", line, 2, true);
						//                    break;
						//                }
						//            }
						//            else
						//                printWarning("\"" + line.Function.ToString() + "\"̏ԈĂ܂", line, 1, true);
						//            break;
						//        }
						//        arg = new SpStrFindArgument(target, word, start);
						//        break;
						//    }
						//case BuiltInFunctionCode.__ARG_SP_FINDCHARA__:
						//    {
						//        TokenReader.SkipWhiteSpace(st);
						//        varToken = VariableParser.ReduceVariable(st);
						//        bool isInteger = varToken.Identifier.IsInteger;
						//        if (varToken == null)
						//        { printWarning("\"" + line.Function.ToString() + "\"̑PFł܂", line, 2, true); break; }
						//        else if (!varToken.Identifier.IsCharacterData)
						//        { printWarning("\"" + line.Function.ToString() + "\"̑PɃLN^ϐȊOw肷邱Ƃ͂ł܂", line, 2, true); break; }
						//        TokenReader.SkipWhiteSpace(st);
						//        if (st.Current != ',')
						//        {
						//            if (st.EOS)
						//                printWarning("\"" + line.Function.ToString() + "\"̈܂", line, 2, true);
						//            else
						//                printWarning("\"" + line.Function.ToString() + "\"̏ԈĂ܂", line, 2, true);
						//            break;
						//        }
						//        st.ShiftNext();
						//        TokenReader.SkipWhiteSpace(st);
						//        IOperandTerm word = ExpressionParser.ReduceExpressionTerm(st);
						//        if (word == null)
						//        { printWarning("\"" + line.Function.ToString() + "\"̑QFł܂", line, 2, true); break; }
						//        else if (isInteger && word.GetOperandType() != typeof(Int64))
						//        { printWarning("\"" + line.Function.ToString() + "\"̈̌^v܂il^̕ϐ𕶎Ō悤ƂĂ܂j", line, 2, true); break; }
						//        else if (!isInteger && word.GetOperandType() != typeof(string))
						//        { printWarning("\"" + line.Function.ToString() + "\"̈̌^v܂i^̕ϐ𐔒lŌ悤ƂĂ܂j", line, 2, true); break; }

						//        TokenReader.SkipWhiteSpace(st);
						//        IOperandTerm start = new SingleTerm(0);
						//        if (!st.EOS)
						//        {
						//            if (st.Current == ',')
						//            {
						//                start = ExpressionParser.ReduceIntegerTerm(st);
						//                if (start == null)
						//                {
						//                    printWarning("\"" + line.Function.ToString() + "\"̑RFł܂", line, 2, true);
						//                    break;
						//                }
						//            }
						//            else
						//                printWarning("\"" + line.Function.ToString() + "\"̏ԈĂ܂", line, 1, true);
						//            break;
						//        }
						//        arg = new SpFindCharaArgument(varToken, word, start);
						//        break;
						//    }
						case BuiltInFunctionCode.__ARG_CASE__:
							{
								CaseExpression[] args = ExpressionParser.ReduceCaseExpressions(st);
								arg = new CaseArgument(args);
								if (!st.EOS)
									printWarning("\"" + line.Function.ToString() + "\"̏ԈĂ܂", line, 1, true);
							}
							break;
                        case BuiltInFunctionCode.__ARG_VAR_INT__:
                            {
                                TokenReader.SkipWhiteSpace(st);
                                varToken = null;
                                if (!st.EOS)
                                    varToken = VariableParser.ReduceVariable(st);
                                if (varToken != null)
                                {
                                    if (varToken.IsString)
                                        printWarning("\"" + line.Function.ToString() + "\"̑P𕶎ϐɂ邱Ƃ͂ł܂", line, 2, true);
                                    else if (varToken.Identifier.Readonly)
                                        printWarning("\"" + line.Function.ToString() + "\"̑PɕύXłȂϐw肷邱Ƃ͂ł܂", line, 2, true);
                                }
                                arg = new PrintDataArgument(varToken);
                            }
                            break;
                        case BuiltInFunctionCode.__ARG_BIT_ARG__:
                            {
                                TokenReader.SkipWhiteSpace(st);
                                varToken = null;
                                varToken = VariableParser.ReduceVariable(st);
                                if (varToken == null)
                                {
                                    printWarning("\"" + line.Function.ToString() + "\"̑PFł܂", line, 2, true);
                                    break;
                                }
                                TokenReader.SkipWhiteSpace(st);
                                if (st.Current != ',')
                                {
                                    printWarning("\"" + line.Function.ToString() + "\"̈܂", line, 2, true);
                                    break;
                                }
                                st.ShiftNext();
                                IOperandTerm y = ExpressionParser.ReduceIntegerTerm(st);
                                TokenReader.SkipWhiteSpace(st);
                                if (st.Current == ',')
                                    printWarning("\"" + line.Function.ToString() + "\"̈܂", line, 1, false);
                                if (varToken.IsString)
                                    printWarning("\"" + line.Function.ToString() + "\"̑P𕶎ϐɂ邱Ƃ͂ł܂", line, 2, true);
                                else if (varToken.Identifier.Readonly)
                                    printWarning("\"" + line.Function.ToString() + "\"̑PɕύXłȂϐw肷邱Ƃ͂ł܂", line, 2, true);

                                if (y is SingleTerm)
                                {
                                    Int64 bit = ((SingleTerm)y).Int;
                                    if ((bit < 0) || (bit > 63))
                                    {
                                        printWarning("\"" + line.Function.ToString() + "\"̑Q͈(O`UR)𒴂Ă܂", line, 2, true);
                                        break;
                                    }
                                }
                                arg = new BitArgument(varToken, y);
                            }
							break;
						case BuiltInFunctionCode.__ARG_SP_VAR_SET__:
							{
								TokenReader.SkipWhiteSpace(st);
								varToken = VariableParser.ReduceVariable(st);
								if (varToken == null)
									printWarning("\"" + line.Function.ToString() + "\"̑PFł܂", line, 2, true);
								else if (varToken.Identifier.Readonly)
									printWarning("\"" + line.Function.ToString() + "\"̑PɕύXłȂϐw肷邱Ƃ͂ł܂", line, 2, true);
								term = null;
								TokenReader.SkipWhiteSpace(st);
								if (st.Current == ',')
								{
									st.ShiftNext();
									term = ExpressionParser.ReduceExpressionTerm(st);
									if (term == null)
										printWarning("\"" + line.Function.ToString() + "\"̑QFł܂", line, 1, false);
								}
								if (term == null)
								{
									if(varToken.IsString)
										term = new SingleTerm("");
									else
										term = new SingleTerm(0);
								}
								else if(term.GetOperandType() != varToken.GetOperandType())
								{
									printWarning("\"" + line.Function.ToString() + "\"̂Q̈̌^vĂ܂", line, 2, true);
									break;
								}
								arg = new SpVarSetArgument(varToken, term);
							}
							break;
						case BuiltInFunctionCode.__ARG_SP_BUTTON__:
						//case BuiltInFunctionCode.__ARG_SP_BUTTONS__:
							{
								IOperandTerm x, y = null;
								TokenReader.SkipWhiteSpace(st);
								x = ExpressionParser.ReduceExpressionTerm(st);
								if (x == null)
									printWarning("\"" + line.Function.ToString() + "\"̑PFł܂", line, 2, true);
								else if (x.GetOperandType() != typeof(string))
									printWarning("\"" + line.Function.ToString() + "\"̑P͕^łȂ΂Ȃ܂", line, 2, true);

								TokenReader.SkipWhiteSpace(st);
								if (st.Current != ',')
								{
									printWarning("\"" + line.Function.ToString() + "\"̑QFł܂", line, 2, true);
									break;
								}
								st.ShiftNext();
								y = ExpressionParser.ReduceExpressionTerm(st);
								if (y == null)
									printWarning("\"" + line.Function.ToString() + "\"̑QFł܂", line, 2, true);
								arg = new SpButtonArgument(x, y);
							}
							break;
						case BuiltInFunctionCode.__ARG_SP_SET_ARRAY__:
						case BuiltInFunctionCode.__ARG_SP_SETS_ARRAY__:
							{
								Type type = typeof(long);
								if (argType == BuiltInFunctionCode.__ARG_SP_SETS_ARRAY__)
									type = typeof(string);
								varToken = VariableParser.ReduceVariable(new StringStream(line.AssignmentDestStr));
								IOperandTerm[] termList = ExpressionParser.ReduceArguments(st, false);
								foreach (IOperandTerm t in termList)
								{
									if ((t == null) || (t.GetOperandType() != type))
									{
										printWarning("Ƒľ^Ⴂ܂", line, 2, true);
										break;
									}
								}
								arg = new SpSetArrayArgument(varToken, termList);
								break;
							}
						default:
							printWarning("ݒ肳ĂȂ^", line, 2, true);
							return false;
					}
				}
			}
			catch (CodeEE e)
			{
                System.Media.SystemSounds.Hand.Play();
                printWarning(e.Message, line, 2, true);
				return false;
			}
			if (arg == null)
			{
                if (!line.IsError)
                {
                    System.Media.SystemSounds.Hand.Play();
                    printWarning("߂̈͒ɓłȂG[", line, 2, true);
                }
				return false;
			}
			line.Argument = arg;
			return true;
		}
	}
}
